"""
Test selectors level 4.

```
[foo='bar' i]
:nth-child(an+b [of S]?)
:is(s1, s2, ...) / :matches(s1, s2, ...)
:where(s1, s2, ...) allowed, but due to our environment, works like `:is()`
:not(s1, s2, ...)
:has(> s1, ...)
```

Likely to be implemented:

- `:nth-col(n)` / `:nth-last-col(n)`: This needs further understanding before implementing and would likely only be
  implemented for HTML, XHTML, and HTML5. This would not be implemented for XML.

- `E || F`: This would need more understanding before implementation. This would likely only be implemented for HTML,
  XHTML, and HTML5. This would not be implemented for XML.

Not supported (with current opinions or plans the matter):

- `:blank`: This applies to inputs with empty or otherwise null input. Currently, there is no plans to implement this.

- `:dir(ltr)`: This applies to direction of text. This direction can be inherited from parents. Due to the way Soup
  Sieve process things, it would have to scan the parents and evaluate what is inherited. it doesn't account for the CSS
  `direction` value, which is a good thing. It is doable, but not sure worth the effort. In addition, it seems there is
  reference to being able to do something like `[dir=auto]` which would select either `ltr` or `rtl`. This seems to add
  additional logic in to attribute selections which would complicate things, but still technically doable. There are
  currently no plans to implement this.

- `:lang(en-*)`: As mentioned in level 2 tests, in documents, `:lang()` can take into considerations information in
  `meta` and other things in the header. At this point, there are no plans to implement this. If a reasonable proposal
  was introduced on how to support this, it may be considered.

- `:local-link`: In our environment, there is no document URL. This isn't currently practical. This will not be
  implemented.

- `:read-only` / `:read-write`: There are no plans to implement this at this time.

- `:required` / `:optional`: There are no plans to implement this at this time.

- `:placeholder-shown`: There are no plans to implement this at this time.

- `:indeterminate`: There are no plans to implement this at this time.

- `:valid` / `:invalid`: We currently to not validate values, so this doesn't make sense at this time.

- `:user-invalid`: User cannot alter things in our environment because there is no user interaction (we are not a
  browser). This will not be implemented.

- `:scope`: I'm not sure what this means or if it is even useful in our context. More information would be needed. It
  seems in an HTML document, this would normally just be `:root` as there is no way to specify a different reference at
  this time. I'm not sure it makes sense to bother implementing this.

- `:in-range` / `:out-of-range`: This applies to form elements only. You'd have to evaluate `value`, `min`, and `max`. I
  guess you can have numerical ranges and alphabetic ranges. Currently, there are no plans to implement this.

- `:current` / `:past` / `:future`: I believe this requires a live, in browser state to determine what is current, to
  then determine what is past and future. I don't think this is applicable to our environment.

- `:default`: This is in the same vain as `:checked`. If we ever implemented that, we'd probably implement this, but
  there are no plans to do so at this time.

- `:focus-within` / `:focus-visible`: There is no focus in our environment, so this will not be implemented.

- `:target-within`: Elements cannot be "targeted" in our environment, so this will not be implemented.

- `:playing` / `:paused`: Elements cannot be played or paused in our environment, so this will not be implemented.
"""
from . import util
import soupsieve as sv


class TestLevel4(util.TestCase):
    """Test level 4 selectors."""

    def test_attribute_case(self):
        """Test attribute value case insensitivity."""

        markup = """
        <div id="div">
        <p id="0" class="somewordshere">Some text <span id="1"> in a paragraph</span>.</p>
        <a id="2" href="http://google.com">Link</a>
        <span id="3" class="herewords">Direct child</span>
        <pre id="pre" class="wordshere">
        <span id="4">Child 1</span>
        <span id="5">Child 2</span>
        <span id="6">Child 3</span>
        </pre>
        </div>
        """

        self.assert_selector(
            markup,
            "[class*=WORDS]",
            [],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            "[class*=WORDS i]",
            ["0", "3", "pre"],
            mode=sv.HTML5
        )

    def test_is_matches_where(self):
        """Test multiple selectors with "is", "matches", and "where"."""

        markup = """
        <div>
        <p>Some text <span id="1"> in a paragraph</span>.
        <a id="2" href="http://google.com">Link</a>
        </p>
        </div>
        """

        self.assert_selector(
            markup,
            ":is(span, a)",
            ["1", "2"],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            ":is(span, a:matches(#2))",
            ["1", "2"],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            ":where(span, a:matches(#2))",
            ["1", "2"],
            mode=sv.HTML5
        )

    def test_multi_nested_not(self):
        """Test nested not and multiple selectors."""

        markup = """
        <div>
        <p id="0">Some text <span id="1"> in a paragraph</span>.</p>
        <a id="2" href="http://google.com">Link</a>
        <span id="3">Direct child</span>
        <pre id="pre">
        <span id="4">Child 1</span>
        <span id="5">Child 2</span>
        <span id="6">Child 3</span>
        </pre>
        </div>
        """

        self.assert_selector(
            markup,
            'div :not(p, :not([id=5]))',
            ['5'],
            mode=sv.HTML5
        )

    def test_has(self):
        """Test has."""

        markup = """
        <div id="0" class="aaaa">
            <p id="1" class="bbbb"></p>
            <p id="2" class="cccc"></p>
            <p id="3" class="dddd"></p>
            <div id="4" class="eeee">
            <div id="5" class="ffff">
            <div id="6" class="gggg">
                <p id="7" class="hhhh"></p>
                <p id="8" class="iiii zzzz"></p>
                <p id="9" class="jjjj"></p>
                <div id="10" class="kkkk">
                    <p id="11" class="llll zzzz"></p>
                </div>
            </div>
            </div>
            </div>
        </div>
        """

        markup2 = """
        <div id="0" class="aaaa">
            <p id="1" class="bbbb"></p>
        </div>
        <div id="2" class="cccc">
            <p id="3" class="dddd"></p>
        </div>
        <div id="4" class="eeee">
            <p id="5" class="ffff"></p>
        </div>
        <div id="6" class="gggg">
            <p id="7" class="hhhh"></p>
        </div>
        <div id="8" class="iiii">
            <p id="9" class="jjjj"></p>
            <span id="10"></span>
        </div>
        """

        self.assert_selector(
            markup,
            'div:not(.aaaa):has(.kkkk > p.llll)',
            ['4', '5', '6'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            'p:has(+ .dddd:has(+ div .jjjj))',
            ['2'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            'p:has(~ .jjjj)',
            ['7', '8'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup2,
            'div:has(> .bbbb, .ffff, .jjjj)',
            ['0', '4', '8'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup2,
            'div:has(> :not(.bbbb, .ffff, .jjjj))',
            ['2', '6', '8'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup2,
            'div:not(:has(> .bbbb, .ffff, .jjjj))',
            ['2', '6'],
            mode=sv.HTML5
        )

    def test_nth_child_of_s(self):
        """Test `nth` child with selector."""

        markup = """
        <p id="0"></p>
        <p id="1"></p>
        <span id="2" class="test"></span>
        <span id="3"></span>
        <span id="4" class="test"></span>
        <span id="5"></span>
        <span id="6" class="test"></span>
        <p id="7"></p>
        <p id="8" class="test"></p>
        <p id="9"></p>
        <p id="10" class="test"></p>
        <span id="11"></span>
        """

        self.assert_selector(
            markup,
            ":nth-child(2n + 1 of :is(p, span).test)",
            ['2', '6', '10'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            ":nth-child(-n+3 of p)",
            ['0', '1', '7'],
            mode=sv.HTML5
        )



"""Utility."""
from collections import Mapping
from collections.abc import Hashable
import bs4
from functools import wraps
import warnings

HTML5 = 0x1
HTML = 0x2
XHTML = 0x4
XML = 0x8

TAG = bs4.Tag
CHILD = (TAG, bs4.Doctype, bs4.Declaration, bs4.CData, bs4.ProcessingInstruction)
COMMENT = bs4.Comment

LC_A = ord('a')
LC_Z = ord('z')
UC_A = ord('A')
UC_Z = ord('Z')


def lower(string):
    """Lower."""

    new_string = []
    for c in string:
        o = ord(c)
        new_string.append(chr(o + 32) if UC_A <= o <= UC_Z else c)
    return ''.join(new_string)


def upper(string):  # pragma: no cover
    """Lower."""

    new_string = []
    for c in string:
        o = ord(c)
        new_string.append(chr(o - 32) if LC_A <= o <= LC_Z else c)
    return ''.join(new_string)


class Immutable:
    """Immutable."""

    __slots__ = ('_hash',)

    def __init__(self, **kwargs):
        """Initialize."""

        temp = []
        for k, v in kwargs.items():
            temp.append(type(v))
            temp.append(v)
            super().__setattr__(k, v)
        super().__setattr__('_hash', hash(tuple(temp)))

    @classmethod
    def __base__(cls):
        """Get base class."""

        return cls

    def __eq__(self, other):
        """Equal."""

        return (
            isinstance(other, self.__base__()) and
            all([getattr(other, key) == getattr(self, key) for key in self.__slots__ if key != '_hash'])
        )

    def __ne__(self, other):
        """Equal."""

        return (
            not isinstance(other, self.__base__()) or
            any([getattr(other, key) != getattr(self, key) for key in self.__slots__ if key != '_hash'])
        )

    def __hash__(self):
        """Hash."""

        return self._hash

    def __setattr__(self, name, value):
        """Prevent mutability."""

        raise AttributeError('Class is immutable!')


class ImmutableDict(Mapping):
    """Hashable, immutable dictionary."""

    def __init__(self, *args, **kwargs):
        """Initialize."""

        if (
            not all([isinstance(v, Hashable) for v in kwargs.values()]) or
            not all([isinstance(v, Hashable) for k, v in args])
        ):
            raise ValueError('All values must be hashable')

        self._d = dict(*args, **kwargs)
        self._hash = hash(tuple([(type(x), x, type(y), y) for x, y in sorted(self._d.items())]))

    def __iter__(self):
        """Iterator."""

        return iter(self._d)

    def __len__(self):
        """Length."""

        return len(self._d)

    def __getitem__(self, key):
        """Get item: `namespace['key']`."""
        return self._d[key]

    def __hash__(self):
        """Hash."""

        return self._hash

    def __repr__(self):
        """Representation."""

        return "%r" % self._d

    __str__ = __repr__


def deprecated(message, stacklevel=2):
    """
    Raise a `DeprecationWarning` when wrapped function/method is called.

    Borrowed from https://stackoverflow.com/a/48632082/866026
    """

    def _decorator(func):
        @wraps(func)
        def _func(*args, **kwargs):
            warnings.warn(
                "'{}' is deprecated. {}".format(func.__name__, message),
                category=DeprecationWarning,
                stacklevel=stacklevel
            )
            return func(*args, **kwargs)
        return _func
    return _decorator



"""
Test selectors level 3.

```
namespace|*
[foo^='bar']
[foo$='bar']
[foo*='bar']
:not(s)
E ~ F
:root
:empty
:last-child
:only-child
:first-of-type
:last-of-type
:only-of-type
:nth-child(an+b)
:nth-last-child(an+b)
:nth-of-type(an+b)
:nth-of-last-type(an+b)
```

Not supported (with current opinions or plans the matter):

- `:target`: Elements cannot be targeted in our environment, so this will not be implemented.

- `:enabled` / `:disabled`: There are currently no plans to implement this.

- `:checked`: Checked could be complicated to implement. Would you select the first item in a `<select>` tag if no
  option is specified to be selected? Or would you consider this something that occurs when the document is live in the
  browser?  There would need to be some considerations. What if a user has selected multiple radio boxes on accident?
  Is this even useful in the context of how Soup Sieve would be used?
"""
from . import util
import soupsieve as sv


class TestLevel3(util.TestCase):
    """Test level 3 selectors."""

    def test_distant_sibling(self):
        """Test distant sibling."""

        self.assert_selector(
            """
            <div>
            <p id="0">Some text <span id="1"> in a paragraph</span>.</p>
            <a id="2" href="http://google.com">Link</a>
            <span id="3">Direct child</span>
            <pre>
            <span id="4">Child 1</span>
            <span id="5">Child 2</span>
            <span id="6">Child 3</span>
            </pre>
            </div>
            """,
            "p ~ span",
            ["3"],
            mode=sv.HTML5
        )

    def test_not(self):
        """Test not."""

        markup = """
        <div>
        <p id="0">Some text <span id="1"> in a paragraph</span>.</p>
        <a id="2" href="http://google.com">Link</a>
        <span id="3">Direct child</span>
        <pre id="pre">
        <span id="4">Child 1</span>
        <span id="5">Child 2</span>
        <span id="6">Child 3</span>
        </pre>
        </div>
        """

        self.assert_selector(
            markup,
            'div :not([id="1"])',
            ["0", "2", "3", "4", "5", "6", "pre"],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            'span:not([id="1"])',
            ["3", "4", "5", "6"],
            mode=sv.HTML5
        )

    def test_attribute_begins(self):
        """Test attribute whose value begins with the specified value."""

        self.assert_selector(
            """
            <div id="div">
            <p id="0" class="herearesomewords">Some text <span id="1"> in a paragraph</span>.</p>
            <a id="2" href="http://google.com">Link</a>
            <span id="3">Direct child</span>
            <pre id="pre">
            <span id="4">Child 1</span>
            <span id="5">Child 2</span>
            <span id="6">Child 3</span>
            </pre>
            </div>
            """,
            "[class^=here]",
            ["0"],
            mode=sv.HTML5
        )

    def test_attribute_end(self):
        """Test attribute whose value ends with the specified value."""

        self.assert_selector(
            """
            <div id="div">
            <p id="0" class="herearesomewords">Some text <span id="1"> in a paragraph</span>.</p>
            <a id="2" href="http://google.com">Link</a>
            <span id="3">Direct child</span>
            <pre id="pre">
            <span id="4">Child 1</span>
            <span id="5">Child 2</span>
            <span id="6">Child 3</span>
            </pre>
            </div>
            """,
            "[class$=words]",
            ["0"],
            mode=sv.HTML5
        )

    def test_attribute_contains(self):
        """Test attribute whose value contains the specified value."""

        markup = """
        <div id="div">
        <p id="0" class="somewordshere">Some text <span id="1"> in a paragraph</span>.</p>
        <a id="2" href="http://google.com">Link</a>
        <span id="3" class="herewords">Direct child</span>
        <pre id="pre" class="wordshere">
        <span id="4">Child 1</span>
        <span id="5">Child 2</span>
        <span id="6">Child 3</span>
        </pre>
        </div>
        """

        self.assert_selector(
            markup,
            "[class*=words]",
            ["0", "3", "pre"],
            mode=sv.HTML5
        )

    def test_root(self):
        """Test root."""

        markup = """
        <html id="root">
        <head>
        </head>
        <body>
        <div id="div">
        <p id="0" class="somewordshere">Some text <span id="1"> in a paragraph</span>.</p>
        <a id="2" href="http://google.com">Link</a>
        <span id="3" class="herewords">Direct child</span>
        <pre id="pre" class="wordshere">
        <span id="4">Child 1</span>
        <span id="5">Child 2</span>
        <span id="6">Child 3</span>
        </pre>
        </div>
        </body>
        </html>
        """

        # Root in HTML is `<html>`
        self.assert_selector(
            markup,
            ":root",
            ["root"],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            ":root > body > div",
            ["div"],
            mode=sv.HTML5
        )

    def test_empty(self):
        """Test empty."""

        markup = """
        <div id="div">
        <p id="0" class="somewordshere">Some text <span id="1"> in a paragraph</span>.</p>
        <a id="2" href="http://google.com">Link</a>
        <span id="3" class="herewords">Direct child</span>
        <pre id="pre" class="wordshere">
        <span id="4"> <!-- comment --> </span>
        <span id="5"> </span>
        <span id="6"></span>
        <span id="7"><span id="8"></span></span>
        </pre>
        </div>
        """

        self.assert_selector(
            markup,
            "body :empty",
            ["4", "5", "6", "8"],
            mode=sv.HTML5
        )

    def test_last_child(self):
        """Test last child."""

        self.assert_selector(
            """
            <div id="div">
            <p id="0">Some text <span id="1"> in a paragraph</span>.</p>
            <a id="2" href="http://google.com">Link</a>
            <span id="3">Direct child</span>
            <pre id="pre">
            <span id="4">Child 1</span>
            <span id="5">Child 2</span>
            <span id="6">Child 3</span>
            </pre>
            </div>
            """,
            "span:last-child",
            ["1", "6"],
            mode=sv.HTML5
        )

    def test_only_child(self):
        """Test only child."""

        self.assert_selector(
            """
            <div id="div">
            <p id="0">Some text <span id="1"> in a paragraph</span>.</p>
            <a id="2" href="http://google.com">Link</a>
            <span id="3">Direct child</span>
            <pre id="pre">
            <span id="4">Child 1</span>
            <span id="5">Child 2</span>
            <span id="6">Child 3</span>
            </pre>
            </div>
            """,
            "span:only-child",
            ["1"],
            mode=sv.HTML5
        )

    def test_namespace(self):
        """Test namespace."""

        markup = """
        <?xml version="1.0" encoding="UTF-8"?>
        <tag id="root">
          <head id="0"></head>
          <foo:other id="1" xmlns:foo="http://me.com/namespaces/foofoo"
                 xmlns:bar="http://me.com/namespaces/foobar">
          <foo:head id="2">
            <foo:title id="3"></foo:title>
            <bar:title id="4"></bar:title>
          </foo:head>
          <body id="5">
            <foo:e1 id="6"></foo:e1>
            <bar:e1 id="7"></bar:e1>
            <e1 id="8"></e1>
            <foo:e2 id="9"></foo:e2>
            <bar:e2 id="10"></bar:e2>
            <e2 id="11"></e2>
            <foo:e3 id="12"></foo:e3>
            <bar:e3 id="13"></bar:e3>
            <e3 id="14"></e3>
          </body>
          </foo:other>
          <other id="15" xmlns="http://me.com/namespaces/other">
            <e4 id="16">Inherit</er>
          </other>
        </tag>
        """

        self.assert_selector(
            markup,
            "foo|title",
            ["3"],
            namespaces={
                "foo": "http://me.com/namespaces/foofoo",
                "bar": "http://me.com/namespaces/foobar"
            },
            mode=sv.XML
        )

        self.assert_selector(
            markup,
            "bar|*",
            ["4", "7", "10", "13"],
            namespaces={
                "foo": "http://me.com/namespaces/foofoo",
                "bar": "http://me.com/namespaces/foobar"
            },
            mode=sv.XML
        )

        self.assert_selector(
            markup,
            "|head",
            ["0"],
            namespaces={
                "foo": "http://me.com/namespaces/foofoo",
                "bar": "http://me.com/namespaces/foobar"
            },
            mode=sv.XML
        )

        self.assert_selector(
            markup,
            "*|e2",
            ["9", "10", "11"],
            namespaces={
                "foo": "http://me.com/namespaces/foofoo",
                "bar": "http://me.com/namespaces/foobar"
            },
            mode=sv.XML
        )

        # Because we employ level 4 selectors
        # E, when no default namespace is defined, will be read as *|E.
        self.assert_selector(
            markup,
            "e3",
            ["12", "13", "14"],
            namespaces={
                "foo": "http://me.com/namespaces/foofoo",
                "bar": "http://me.com/namespaces/foobar"
            },
            mode=sv.XML
        )

        # Now that we apply a default namespace. Null space.
        self.assert_selector(
            markup,
            "e3",
            ["14"],
            namespaces={
                "": "",
                "foo": "http://me.com/namespaces/foofoo",
                "bar": "http://me.com/namespaces/foobar"
            },
            mode=sv.XML
        )

        self.assert_selector(
            markup,
            "head",
            ["0"],
            namespaces={
                "": "",
                "foo": "http://me.com/namespaces/foofoo",
                "bar": "http://me.com/namespaces/foobar"
            },
            mode=sv.XML
        )

        # Because no prefix is specified for "other" in the above document,
        # `e4` inherits the other namespace. The prefix in this case doesn't matter.
        # We specify `other` as prefix in our CSS just so we can use it to target the element.
        self.assert_selector(
            markup,
            "e4",
            [],
            namespaces={
                "": "",
                "foo": "http://me.com/namespaces/foofoo",
                "bar": "http://me.com/namespaces/foobar",
                "other": "http://me.com/namespaces/other"
            },
            mode=sv.XML
        )

    def test_attribute_namespace(self):
        """Test attribute namespace."""

        markup = """
        <!DOCTYPE html>
        <html>
        <head>
          <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
        </head>
        <body>
          <h1>A contrived example</h1>
          <svg viewBox="0 0 20 32" class="icon icon-1">
            <use id="0" xlink:href="images/sprites.svg#icon-undo">aaaa</use>
          </svg>
          <svg viewBox="0 0 30 32" class="icon icon-2">
            <use id="1" xlink:href="images/sprites.svg#icon-redo">bbbb</use>
          </svg>
          <svg viewBox="0 0 40 32" class="icon icon-3">
            <use id="2" xlink:href="images/sprites.svg#icon-forward">cccc</use>
          </svg>
          <svg viewBox="0 0 50 32" class="icon icon-4">
            <use id="3" xlink:href="other/sprites.svg#icon-reply">dddd</use>
          </svg>
        </body>
        </html>
        """

        self.assert_selector(
            markup,
            '[xlink|href*=forw],[xlink|href="images/sprites.svg#icon-redo"]',
            ['1', '2'],
            namespaces={"xlink": "http://www.w3.org/1999/xlink"},
            mode=sv.HTML5
        )

    def test_first_of_type(self):
        """Test first of type."""

        markup = """
        <p id="0"></p>
        <p id="1"></p>
        <span id="2"></span>
        <span id="3"></span>
        <span id="4"></span>
        <span id="5"></span>
        <span id="6"></span>
        <p id="7"></p>
        <p id="8"></p>
        <p id="9"></p>
        <p id="10"></p>
        <span id="11"></span>
        """

        self.assert_selector(
            markup,
            "p:first-of-type",
            ['0'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            "span:first-of-type",
            ['2'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            "body :first-of-type",
            ['0', '2'],
            mode=sv.HTML5
        )

    def test_last_of_type(self):
        """Test last of type."""

        markup = """
        <p id="0"></p>
        <p id="1"></p>
        <span id="2"></span>
        <span id="3"></span>
        <span id="4"></span>
        <span id="5"></span>
        <span id="6"></span>
        <p id="7"></p>
        <p id="8"></p>
        <p id="9"></p>
        <p id="10"></p>
        <span id="11"></span>
        """

        self.assert_selector(
            markup,
            "p:last-of-type",
            ['10'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            "span:last-of-type",
            ['11'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            "body :last-of-type",
            ['10', '11'],
            mode=sv.HTML5
        )

    def test_only_of_type(self):
        """Test only of type."""

        markup = """
        <div id="0">
            <p id="1"></p>
        </div>
        <span id="2"></span>
        <span id="3"></span>
        <p id="4"></p>
        <span id="5"></span>
        <span id="6"></span>
        <div>
            <p id="7"></p>
            <p id="8"></p>
        </div>
        """

        self.assert_selector(
            markup,
            "p:only-of-type",
            ['1', '4'],
            mode=sv.HTML5
        )

    def test_nth_child(self):
        """Test `nth` child."""

        markup = """
        <p id="0"></p>
        <p id="1"></p>
        <span id="2"></span>
        <span id="3"></span>
        <span id="4"></span>
        <span id="5"></span>
        <span id="6"></span>
        <p id="7"></p>
        <p id="8"></p>
        <p id="9"></p>
        <p id="10"></p>
        <span id="11"></span>
        """

        self.assert_selector(
            markup,
            "p:nth-child(-2)",
            [],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            "p:nth-child(2)",
            ['1'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            "p:nth-child(9n - 1)",
            ['7'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            "p:nth-child(2n + 1)",
            ['0', '8', '10'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            "p:nth-child(-n+3)",
            ['0', '1'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            "span:nth-child(-n+3)",
            ['2'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            "body *:nth-child(-n+3)",
            ['0', '1', '2'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            "p:nth-child(odd)",
            ['0', '8', '10'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            "p:nth-child(even)",
            ['1', '7', '9'],
            mode=sv.HTML5
        )

    def test_nth_last_child(self):
        """Test `nth` last child."""

        markup = """
        <p id="0"></p>
        <p id="1"></p>
        <span id="2"></span>
        <span id="3"></span>
        <span id="4"></span>
        <span id="5"></span>
        <span id="6"></span>
        <p id="7"></p>
        <p id="8"></p>
        <p id="9"></p>
        <p id="10"></p>
        <span id="11"></span>
        """

        self.assert_selector(
            markup,
            "p:nth-last-child(2)",
            ['10'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            "p:nth-last-child(2n + 1)",
            ['1', '7', '9'],
            mode=sv.HTML5
        )

    def test_nth_of_type(self):
        """Test `nth` of type."""

        markup = """
        <p id="0"></p>
        <p id="1"></p>
        <span id="2"></span>
        <span id="3"></span>
        <span id="4"></span>
        <span id="5"></span>
        <span id="6"></span>
        <p id="7"></p>
        <p id="8"></p>
        <p id="9"></p>
        <p id="10"></p>
        <span id="11"></span>
        """

        self.assert_selector(
            markup,
            "p:nth-of-type(3)",
            ['7'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            "p:nth-of-type(2n + 1)",
            ['0', '7', '9'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            "span:nth-of-type(2n + 1)",
            ['2', '4', '6'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            "body :nth-of-type(2n + 1)",
            ['0', '2', '4', '6', '7', '9'],
            mode=sv.HTML5
        )

    def test_nth_last_of_type(self):
        """Test `nth` last of type."""

        markup = """
        <p id="0"></p>
        <p id="1"></p>
        <span id="2"></span>
        <span id="3"></span>
        <span id="4"></span>
        <span id="5"></span>
        <span id="6"></span>
        <p id="7"></p>
        <p id="8"></p>
        <p id="9"></p>
        <p id="10"></p>
        <span id="11"></span>
        """

        self.assert_selector(
            markup,
            "p:nth-last-of-type(3)",
            ['8'],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            "p:nth-last-of-type(2n + 1)",
            ['1', '8', '10'],
            mode=sv.HTML5
        )



"""CSS selector parser."""
import re
import copyreg
from collections import namedtuple
from functools import lru_cache
from . import util
from . import css_match as cm
from .util import deprecated


# Selector patterns
CSS_ESCAPES = r'(?:\\[a-fA-F0-9]{1,6}[ ]?|\\.)'
NTH = r'(?:[-+])?(?:\d+n?|n)(?:(?<=n)\s*(?:[-+])\s*(?:\d+))?'
HTML_SELECTORS = r"""
(?P<class_id>(?:\#|\.)(?:[-\w]|{esc})+) |                            #.class and #id
(?P<ns_tag>(?:(?:(?:[-\w]|{esc})+|\*)?\|)?(?:(?:[-\w]|{esc})+|\*)) | # namespace:tag
\[(?P<ns_attr>(?:(?:(?:[-\w]|{esc})+|\*)?\|)?(?:[-\w]|{esc})+)
""".format(esc=CSS_ESCAPES)
XML_SELECTORS = r"""
(?P<ns_tag>(?:(?:(?:[-\w.]|{esc})+|\*)?\|)?(?:(?:[-\w.]|{esc})+|\*)) | # namespace:tag
\[(?P<ns_attr>(?:(?:(?:[-\w]|{esc})+|\*)\|)?(?:[-\w.]|{esc})+)         # namespace:attributes
""".format(esc=CSS_ESCAPES)
SELECTORS = r'''(?x)
    (?P<pseudo_open>:(?:has|is|matches|not|where)\() |                            # optinal pseudo selector wrapper
    (?P<pseudo>:(?:empty|root|(?:first|last|only)-(?:child|of-type))) |           # Simple pseudo selector
    (?P<pseudo_nth_child>:nth-(?:last-)?child
        \(\s*(?P<nth_child>{nth}|even|odd)\s*(?:\)|(?<=\s)of\s+)) |               # Pseudo `nth-child` selectors
    (?P<pseudo_nth_type>:nth-(?:last-)?of-type
        \(\s*(?P<nth_type>{nth}|even|odd)\s*\)) |                                 # Pseudo `nth-of-type` selectors
    {doc_specific}
    (?:(?P<cmp>[~^|*$]?=)                                                         # compare
    (?P<value>"(\\.|[^\\"]+)*?"|'(\\.|[^\\']+)*?'|(?:[^'"\[\] \t\r\n]|{esc})+))?  # attribute value
    (?P<i>[ ]+i)? \] |                                                            # case insensitive
    (?P<pseudo_close>\)) |                                                        # optional pseudo selector close
    (?P<split>\s*?(?P<relation>[,+>~]|[ ](?![,+>~]))\s*) |                        # split multiple selectors
    (?P<invalid>).+                                                               # not proper syntax
'''
RE_HTML_SEL = re.compile(SELECTORS.format(esc=CSS_ESCAPES, nth=NTH, doc_specific=HTML_SELECTORS))
RE_XML_SEL = re.compile(SELECTORS.format(esc=CSS_ESCAPES, nth=NTH, doc_specific=XML_SELECTORS))

# CSS escape pattern
RE_CSS_ESC = re.compile(r'(?:(\\[a-fA-F0-9]{1,6}[ ]?)|(\\.))')

# Pattern to break up `nth` specifiers
RE_NTH = re.compile(r'(?P<s1>[-+])?(?P<a>\d+n?|n)(?:(?<=n)\s*(?P<s2>[-+])\s*(?P<b>\d+))?')

SPLIT = ','
HAS_CHILD = ": "

_MAXCACHE = 500


@lru_cache(maxsize=_MAXCACHE)
def _cached_css_compile(pattern, namespaces, mode):
    """Cached CSS compile."""

    return SoupSieve(
        pattern,
        namespaces,
        mode
    )


def _purge_cache():
    """Purge the cache."""

    _cached_css_compile.cache_clear()


def css_unescape(string):
    """Unescape CSS value."""

    def replace(m):
        """Replace with the appropriate substitute."""

        return chr(int(m.group(1)[1:], 16)) if m.group(1) else m.group(2)[1:]

    return RE_CSS_ESC.sub(replace, string)


class _Namespaces(util.ImmutableDict):
    """Namespaces."""

    def __init__(self, *args, **kwargs):
        """Initialize."""

        if not all([isinstance(k, str) and isinstance(v, str) for k, v in args]):
            raise TypeError('Namespace keys and values must be Unicode strings')
        if not all([isinstance(k, str) and isinstance(v, str) for k, v in kwargs.items()]):
            raise TypeError('Namespace keys and values must be Unicode strings')

        super().__init__(*args, **kwargs)


class _Selector:
    """Intermediate selector."""

    def __init__(self, **kwargs):
        """Initialize."""

        self.tags = kwargs.get('tags', [])
        self.ids = kwargs.get('ids', [])
        self.classes = kwargs.get('classes', [])
        self.attributes = kwargs.get('attributes', [])
        self.nth = kwargs.get('nth', [])
        self.selectors = kwargs.get('selectors', [])
        self.is_not = kwargs.get('is_not', False)
        self.is_empty = kwargs.get('is_empty', False)
        self.relation = kwargs.get('relation', None)
        self.rel_type = kwargs.get('rel_type', None)
        self.is_root = kwargs.get('is_root', False)

    def set_distant_relation(self, value):
        """Set the furthest relation down the chain."""

        if self.relation:
            temp = self.relation
            while temp and temp.relation:
                temp = temp.relation
            temp.relation = value
        else:
            self.relation = value

    def _freeze_list(self, chain):
        """Create an immutable selector chain."""

        for index, item in enumerate(chain):
            if isinstance(item, list):
                chain[index] = self._freeze_list(item)
            elif isinstance(item, _Selector):
                chain[index] = item.freeze()
        return tuple(chain)

    def freeze(self):
        """Freeze self."""

        return Selector(
            self._freeze_list(self.tags),
            self._freeze_list(self.ids),
            self._freeze_list(self.classes),
            self._freeze_list(self.attributes),
            self._freeze_list(self.nth),
            self._freeze_list(self.selectors),
            self.is_not,
            self.is_empty,
            self.relation.freeze() if self.relation is not None else None,
            self.rel_type,
            self.is_root
        )

    def __str__(self):
        """String representation."""

        return (
            '_Selector(tags=%r, ids=%r, classes=%r, attributes=%r, nth=%r, selectors=%r, '
            'is_not=%r, relation=%r, rel_type=%r, is_root=%r)'
        ) % (
            self.tags, self.ids, self.classes, self.attributes, self.nth, self.selectors,
            self.is_not, self.relation, self.rel_type, self.is_root
        )

    __repr__ = __str__


class Selector(
    namedtuple(
        'Selector',
        [
            'tags', 'ids', 'classes', 'attributes', 'nth', 'selectors',
            'is_not', 'is_empty', 'relation', 'rel_type', 'is_root'
        ]
    )
):
    """Selector."""


class SelectorTag(namedtuple('SelectorTag', ['name', 'prefix'])):
    """Selector tag."""


class SelectorAttribute(namedtuple('AttrRule', ['attribute', 'prefix', 'pattern'])):
    """Selector attribute rule."""


class SelectorNth(namedtuple('SelectorNth', ['a', 'n', 'b', 'type', 'last', 'selectors'])):
    """Selector nth type."""


class CSSParser:
    """Parse CSS selectors."""

    def __init__(self, selector, mode=util.HTML5):
        """Initialize."""

        self.pattern = selector

        if mode in (util.HTML, util.HTML5, util.XML, util.XHTML):
            self.mode = mode
        else:
            raise ValueError("Invalid SelectorMatcher flag(s) '{}'".format(mode))
        self.re_sel = RE_HTML_SEL if self.mode != util.XML else RE_XML_SEL

    def parse_attribute_selector(self, sel, m, has_selector):
        """Create attribute selector from the returned regex match."""

        flags = re.I if m.group('i') else 0
        parts = [css_unescape(a.strip()) for a in m.group('ns_attr').split('|')]
        ns = ''
        if len(parts) > 1:
            ns = parts[0]
            attr = parts[1]
        else:
            attr = parts[0]
        op = m.group('cmp')
        if op:
            value = css_unescape(
                m.group('value')[1:-1] if m.group('value').startswith(('"', "'")) else m.group('value')
            )
        else:
            value = None
        if not op:
            # Attribute name
            pattern = None
        elif op.startswith('^'):
            # Value start with
            pattern = re.compile(r'^%s.*' % re.escape(value), flags)
        elif op.startswith('$'):
            # Value ends with
            pattern = re.compile(r'.*?%s$' % re.escape(value), flags)
        elif op.startswith('*'):
            # Value contains
            pattern = re.compile(r'.*?%s.*' % re.escape(value), flags)
        elif op.startswith('~'):
            # Value contains word within space separated list
            pattern = re.compile(r'.*?(?:(?<=^)|(?<= ))%s(?=(?:[ ]|$)).*' % re.escape(value), flags)
        elif op.startswith('|'):
            # Value starts with word in dash separated list
            pattern = re.compile(r'^%s(?:-.*)?$' % re.escape(value), flags)
        else:
            # Value matches
            pattern = re.compile(r'^%s$' % re.escape(value), flags)
        has_selector = True
        sel.attributes.append(SelectorAttribute(attr, ns, pattern))
        return has_selector

    def parse_tag_pattern(self, sel, m, has_selector):
        """Parse tag pattern from regex match."""

        parts = [css_unescape(x) for x in m.group('ns_tag').split('|')]
        if len(parts) > 1:
            prefix = parts[0]
            tag = parts[1]
        else:
            tag = parts[0]
            prefix = None
        sel.tags.append(SelectorTag(tag, prefix))
        has_selector = True
        return has_selector

    def parse_pseudo(self, sel, m, has_selector):
        """Parse pseudo."""

        pseudo = m.group('pseudo')[1:]
        if pseudo == 'root':
            sel.is_root = True
        elif pseudo == 'empty':
            sel.is_empty = True
        elif pseudo == 'first-child':
            sel.nth.append(SelectorNth(1, False, 0, False, False, tuple()))
        elif pseudo == 'last-child':
            sel.nth.append(SelectorNth(1, False, 0, False, True, tuple()))
        elif pseudo == 'first-of-type':
            sel.nth.append(SelectorNth(1, False, 0, True, False, tuple()))
        elif pseudo == 'last-of-type':
            sel.nth.append(SelectorNth(1, False, 0, True, True, tuple()))
        elif pseudo == 'only-child':
            sel.nth.extend(
                [
                    SelectorNth(1, False, 0, False, False, tuple()),
                    SelectorNth(1, False, 0, False, True, tuple())
                ]
            )
        elif pseudo == 'only-of-type':
            sel.nth.extend(
                [
                    SelectorNth(1, False, 0, True, False, tuple()),
                    SelectorNth(1, False, 0, True, True, tuple())
                ]
            )

        has_selector = True
        return has_selector

    def parse_pseudo_nth(self, sel, m, has_selector, iselector):
        """Parse `nth` pseudo."""

        postfix = '_child' if m.group('pseudo_nth_child') else '_type'
        content = m.group('nth' + postfix)
        if content == 'even':
            s1 = 2
            s2 = 2
            var = True
        elif content == 'odd':
            s1 = 2
            s2 = 1
            var = True
        else:
            nth_parts = RE_NTH.match(content)
            s1 = '-' if nth_parts.group('s1') and nth_parts.group('s1') == '-' else ''
            a = nth_parts.group('a')
            var = a.endswith('n')
            if a.startswith('n'):
                s1 += '1'
            elif var:
                s1 += a[:-1]
            else:
                s1 += a
            s2 = '-' if nth_parts.group('s2') and nth_parts.group('s2') == '-' else ''
            if nth_parts.group('b'):
                s2 += nth_parts.group('b')
            else:
                s2 = '0'
            s1 = int(s1, 16)
            s2 = int(s2, 16)

        if postfix == '_child':
            if m.group('pseudo_nth' + postfix).strip().endswith('of'):
                # Parse the rest of `of S`.
                temp_sel = iselector
            else:
                # Use default `*|*` for `of S`. Simulate un-closed pseudo.
                temp_sel = self.re_sel.finditer('*|*)')
            nth_sel = tuple(
                self.parse_selectors(
                    temp_sel,
                    True,
                    False,
                    False
                )
            )
            if m.group('pseudo_nth' + postfix).startswith(':nth-child'):
                sel.nth.append(SelectorNth(s1, var, s2, False, False, nth_sel))
            elif m.group('pseudo_nth' + postfix).startswith(':nth-last-child'):
                sel.nth.append(SelectorNth(s1, var, s2, False, True, nth_sel))
        else:
            if m.group('pseudo_nth' + postfix).startswith(':nth-of-type'):
                sel.nth.append(SelectorNth(s1, var, s2, True, False, tuple()))
            elif m.group('pseudo_nth' + postfix).startswith(':nth-last-of-type'):
                sel.nth.append(SelectorNth(s1, var, s2, True, True, tuple()))
        has_selector = True
        return has_selector

    def parse_pseudo_open(self, sel, m, has_selector, iselector, is_pseudo):
        """Parse pseudo with opening bracket."""

        sel.selectors.extend(
            self.parse_selectors(
                iselector,
                True,
                m.group('pseudo_open')[1:-1] == 'not',
                m.group('pseudo_open')[1:-1] == 'has'
            )
        )
        has_selector = True
        return has_selector

    def parse_has_split(self, sel, m, has_selector, selectors, is_pseudo, rel_type):
        """Parse splitting tokens."""

        if m.group('relation') == SPLIT:
            if not has_selector:
                raise ValueError("Cannot start or end selector with '{}'".format(m.group('relation')))
            sel.rel_type = rel_type
            selectors[-1].set_distant_relation(sel)
            rel_type = HAS_CHILD
            selectors.append(_Selector())
        else:
            if has_selector:
                sel.rel_type = rel_type
                selectors[-1].set_distant_relation(sel)
            rel_type = ':' + m.group('relation')
        sel = _Selector()

        has_selector = False
        return has_selector, sel, rel_type

    def parse_split(self, sel, m, has_selector, selectors, relations, is_pseudo):
        """Parse splitting tokens."""

        if not has_selector:
            raise ValueError("Cannot start or end selector with '{}'".format(m.group('relation')))
        is_not = sel.is_not
        if m.group('relation') == SPLIT:
            if not sel.tags and not is_pseudo:
                # Implied `*`
                sel.tags.append(SelectorTag('*', None))
            sel.relation = relations[0] if relations else None
            selectors.append(sel)
            relations.clear()
        else:
            # In this particular case, we are attaching a relation to an element of interest.
            # the `:not()` applies to the element of interest, not the ancestors. `:not()` is really only
            # applied at the sub-selector level `_Selector(is_not=False, selector=[_Selector(is_not=True)])`
            # We want to see if the element has the proper ancestry, and then apply the `:not()`.
            # In the case of `:not(el1) > el2`, where the ancestor is evaluated with a not, you'd actually get:
            # ```
            # _Selector(
            #     tags=['el2'],
            #     relations=[
            #         [
            #             _Selector(
            #                 tags=['*'],
            #                 selectors=[
            #                     _Selector(
            #                         tags=['el1'],
            #                         is_not=True
            #                     )
            #                 ]
            #             )
            #         ]
            #     ],
            #     rel_type='>'
            # )
            # ```
            sel.relation = relations[0] if relations else None
            sel.rel_type = m.group('relation')
            sel.is_not = False
            relations.clear()
            relations.append(sel)
        sel = _Selector(is_not=is_not)

        has_selector = False
        return has_selector, sel

    def parse_classes(self, sel, m, has_selector):
        """Parse classes."""

        selector = m.group('class_id')
        if selector.startswith('.'):
            sel.classes.append(css_unescape(selector[1:]))
            has_selector = True
        else:
            sel.ids.append(css_unescape(selector[1:]))
            has_selector = True
        return has_selector

    def parse_selectors(self, iselector, is_pseudo=False, is_not=False, is_has=False):
        """Parse selectors."""

        sel = _Selector(is_not=is_not)
        selectors = []
        has_selector = False
        closed = False
        is_html = self.mode != util.XML
        relations = []
        rel_type = HAS_CHILD
        split_last = False
        if is_has:
            selectors.append(_Selector())

        try:
            while True:
                m = next(iselector)

                # Handle parts
                if m.group('pseudo'):
                    has_selector = self.parse_pseudo(sel, m, has_selector)
                elif m.group('pseudo_nth_type') or m.group('pseudo_nth_child'):
                    has_selector = self.parse_pseudo_nth(sel, m, has_selector, iselector)
                elif m.group('pseudo_open'):
                    has_selector = self.parse_pseudo_open(sel, m, has_selector, iselector, is_pseudo)
                elif m.group('pseudo_close'):
                    if split_last:
                        raise ValueError("Cannot end with a combining character")
                    if is_pseudo:
                        closed = True
                        break
                    else:
                        raise ValueError("Bad selector '{}'".format(m.group(0)))
                elif m.group('split'):
                    if split_last:
                        raise ValueError("Cannot have combining character after a combining character")
                    if is_has:
                        has_selector, sel, rel_type = self.parse_has_split(
                            sel, m, has_selector, selectors, is_pseudo, rel_type
                        )
                    else:
                        has_selector, sel = self.parse_split(sel, m, has_selector, selectors, relations, is_pseudo)
                    split_last = True
                    continue
                elif m.group('ns_attr'):
                    has_selector = self.parse_attribute_selector(sel, m, has_selector)
                elif m.group('ns_tag'):
                    if has_selector:
                        raise ValueError("Tag must come first")
                    has_selector = self.parse_tag_pattern(sel, m, has_selector)
                elif is_html and m.group('class_id'):
                    has_selector = self.parse_classes(sel, m, has_selector)
                elif m.group('invalid'):
                    raise ValueError("Bad selector '{}'".format(m.group(0)))
                split_last = False
        except StopIteration:
            pass

        if is_pseudo and not closed:
            raise ValueError("Unclosed `:pseudo()`")

        if split_last:
            raise ValueError("Cannot end with a combining character")

        if has_selector:
            if not sel.tags and not is_pseudo:
                # Implied `*`
                sel.tags.append(SelectorTag('*', None))
            if is_has:
                sel.rel_type = rel_type
                selectors[-1].set_distant_relation(sel)
            else:
                sel.relation = relations[0] if relations else None
                relations.clear()
                selectors.append(sel)
        elif is_has:
            # We will always need to finish a selector when `:has()` is used as it leads with combining.
            raise ValueError('Missing selectors after combining type.')

        return selectors

    def process_selectors(self):
        """
        Process selectors.

        We do our own selectors as BeautifulSoup4 has some annoying quirks,
        and we don't really need to do nth selectors or siblings or
        descendants etc.
        """

        return tuple([s.freeze() for s in self.parse_selectors(self.re_sel.finditer(self.pattern))])


class SoupSieve(util.Immutable):
    """Match tags in Beautiful Soup with CSS selectors."""

    __slots__ = ("pattern", "selectors", "namespaces", "mode", "_hash")

    def __init__(self, selectors, namespaces, mode):
        """Initialize."""

        super().__init__(
            pattern=selectors,
            selectors=CSSParser(selectors, mode).process_selectors(),
            namespaces=namespaces,
            mode=mode
        )

    def _walk(self, node, capture=True, comments=False):
        """Recursively return selected tags."""

        if capture and self.match(node):
            yield node

        # Walk children
        for child in node.descendants:
            if capture and isinstance(child, util.TAG) and self.match(child):
                yield child
            elif comments and isinstance(child, util.COMMENT):
                yield child

    def _sieve(self, node, capture=True, comments=False, limit=0):
        """Sieve."""

        if limit < 1:
            limit = None

        for child in self._walk(node, capture, comments):
            yield child
            if limit is not None:
                limit -= 1
                if limit < 1:
                    break

    def match(self, node):
        """Match."""

        return cm.CSSMatch(self.selectors, self.namespaces, self.mode).match(node)

    def filter(self, nodes):  # noqa A001
        """Filter."""

        if isinstance(nodes, util.TAG):
            return [node for node in nodes.children if isinstance(node, util.TAG) and self.match(node)]
        else:
            return [node for node in nodes if self.match(node)]

    def comments(self, node, limit=0):
        """Get comments only."""

        return list(self.icomments(node, limit))

    def icomments(self, node, limit=0):
        """Iterate comments only."""

        yield from self._sieve(node, capture=False, comments=True, limit=limit)

    def select(self, node, limit=0):
        """Select the specified tags."""

        return list(self.iselect(node, limit))

    def iselect(self, node, limit=0):
        """Iterate the specified tags."""

        yield from self._sieve(node, limit=limit)

    def __repr__(self):
        """Representation."""

        return "SoupSieve(pattern=%r, namespaces=%s, mode=%s)" % (self.pattern, self.namespaces, self.mode)

    __str__ = __repr__

    # ====== Deprecated ======
    @deprecated("Use 'SoupSieve.icomments' instead.")
    def commentsiter(self, node, limit=0):
        """Iterate comments only."""

        yield from self.icomments(node, limit)

    @deprecated("Use 'SoupSieve.iselect' instead.")
    def selectiter(self, node, limit=0):
        """Iterate the specified tags."""

        yield from self.iselect(node, limit)


def _pickle(p):
    return SoupSieve, (p.pattern, p.namespaces, p.mode)


copyreg.pickle(SoupSieve, _pickle)



"""Test Soup Sieve API."""
import unittest
import bs4
import soupsieve as sv
import copy
import random
import pytest
import warnings


class TestSoupSieve(unittest.TestCase):
    """Test Soup Sieve."""

    def test_comments(self):
        """Test comments."""

        markup = """
        <!-- before header -->
        <html>
        <head>
        </head>
        <body>
        <!-- comment -->
        <p id="1"><code id="2"></code><img id="3" src="./image.png"/></p>
        <pre id="4"></pre>
        <p><span id="5" class="some-class"></span><span id="some-id"></span></p>
        <pre id="6" class='ignore'>
            <!-- don't ignore -->
        </pre>
        </body>
        </html>
        """

        soup = bs4.BeautifulSoup(markup, 'html5lib')
        comments = [str(c).strip() for c in sv.comments(soup, mode=sv.HTML5)]
        self.assertEqual(sorted(comments), sorted(['before header', 'comment', "don't ignore"]))

        comments = [str(c).strip() for c in sv.icomments(soup, limit=2, mode=sv.HTML5)]
        self.assertEqual(sorted(comments), sorted(['before header', 'comment']))

    def test_select(self):
        """Test select."""

        markup = """
        <!-- before header -->
        <html>
        <head>
        </head>
        <body>
        <!-- comment -->
        <p id="1"><code id="2"></code><img id="3" src="./image.png"/></p>
        <pre id="4"></pre>
        <p><span id="5" class="some-class"></span><span id="some-id"></span></p>
        <pre id="6" class='ignore'>
            <!-- don't ignore -->
        </pre>
        </body>
        </html>
        """

        soup = bs4.BeautifulSoup(markup, 'html5lib')
        ids = []
        for el in sv.select('span[id]', soup):
            ids.append(el.attrs['id'])

        self.assertEqual(sorted(['5', 'some-id']), sorted(ids))

        ids = []
        for el in sv.iselect('span[id]', soup):
            ids.append(el.attrs['id'])

        self.assertEqual(sorted(['5', 'some-id']), sorted(ids))

    def test_match(self):
        """Test matching."""

        markup = """
        <!-- before header -->
        <html>
        <head>
        </head>
        <body>
        <!-- comment -->
        <p id="1"><code id="2"></code><img id="3" src="./image.png"/></p>
        <pre id="4"></pre>
        <p><span id="5" class="some-class"></span><span id="some-id"></span></p>
        <pre id="6" class='ignore'>
            <!-- don't ignore -->
        </pre>
        </body>
        </html>
        """

        soup = bs4.BeautifulSoup(markup, 'html5lib')
        nodes = sv.select('span[id]', soup)
        self.assertTrue(sv.match('span#5', nodes[0]))
        self.assertFalse(sv.match('span#5', nodes[1]))

    def test_filter(self):
        """Test filter."""

        markup = """
        <!-- before header -->
        <html>
        <head>
        </head>
        <body>
        <!-- comment -->
        <p id="1"><code id="2"></code><img id="3" src="./image.png"/></p>
        <pre id="4"></pre>
        <p><span id="5" class="some-class"></span><span id="some-id"></span></p>
        <pre id="6" class='ignore'>
            <!-- don't ignore -->
        </pre>
        </body>
        </html>
        """

        soup = bs4.BeautifulSoup(markup, 'html5lib')
        nodes = sv.filter('pre#6', soup.html.body)
        self.assertEqual(len(nodes), 1)
        self.assertEqual(nodes[0].attrs['id'], '6')

    def test_cache(self):
        """Test copy."""

        p1 = sv.compile('p[id]', mode=sv.HTML5)
        p2 = sv.compile('p[id]', mode=sv.HTML5)
        p3 = sv.compile('p[id]', mode=sv.HTML)
        p4 = copy.copy(p1)
        p5 = copy.copy(p3)

        self.assertTrue(p1 is p2)
        self.assertTrue(p1 is not p3)
        self.assertTrue(p4 is not p1)
        self.assertTrue(p4 == p1)
        self.assertTrue(p5 is not p3)
        self.assertTrue(p5 == p3)
        self.assertTrue(p5 is not p4)

        sv.purge()
        self.assertEqual(sv.cp._cached_css_compile.cache_info().currsize, 0)
        for x in range(1000):
            value = '[value={}]'.format(str(random.randint(1, 10000)))
            p = sv.compile(value)
            self.assertTrue(p.pattern == value)
            self.assertTrue(sv.cp._cached_css_compile.cache_info().currsize > 0)
        self.assertTrue(sv.cp._cached_css_compile.cache_info().currsize == 500)
        sv.purge()
        self.assertEqual(sv.cp._cached_css_compile.cache_info().currsize, 0)

    def test_recompile(self):
        """If you feed through the same object, it should pass through unless you change flags."""

        p1 = sv.compile('p[id]')
        p2 = sv.compile(p1)
        self.assertTrue(p1 is p2)

        with pytest.raises(ValueError):
            sv.compile(p1, mode=sv.HTML)

        with pytest.raises(ValueError):
            sv.compile(p1, namespaces={"": ""})


class TestDeprcations(unittest.TestCase):
    """Test Soup Sieve deprecations."""

    def test_selectiter_deprecation(self):
        """Test the deprecated iterator functions."""

        markup = """
        <!-- before header -->
        <html>
        <head>
        </head>
        <body>
        <!-- comment -->
        <p id="1"><code id="2"></code><img id="3" src="./image.png"/></p>
        <pre id="4"></pre>
        <p><span id="5" class="some-class"></span><span id="some-id"></span></p>
        <pre id="6" class='ignore'>
            <!-- don't ignore -->
        </pre>
        </body>
        </html>
        """

        soup = bs4.BeautifulSoup(markup, 'html5lib')

        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")

            ids = []
            for el in sv.selectiter('span[id]', soup):
                ids.append(el.attrs['id'])
            self.assertEqual(sorted(['5', 'some-id']), sorted(ids))
            self.assertTrue(len(w) == 1)
            self.assertTrue(issubclass(w[-1].category, DeprecationWarning))

        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")

            ids = []
            for el in sv.compile('span[id]').selectiter(soup):
                ids.append(el.attrs['id'])
            self.assertEqual(sorted(['5', 'some-id']), sorted(ids))
            self.assertTrue(len(w) == 1)
            self.assertTrue(issubclass(w[-1].category, DeprecationWarning))

    def test_commentsiter_deprecation(self):
        """Test the deprecated iterator functions."""

        markup = """
        <!-- before header -->
        <html>
        <head>
        </head>
        <body>
        <!-- comment -->
        <p id="1"><code id="2"></code><img id="3" src="./image.png"/></p>
        <pre id="4"></pre>
        <p><span id="5" class="some-class"></span><span id="some-id"></span></p>
        <pre id="6" class='ignore'>
            <!-- don't ignore -->
        </pre>
        </body>
        </html>
        """

        soup = bs4.BeautifulSoup(markup, 'html5lib')

        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")

            comments = [str(c).strip() for c in sv.commentsiter(soup, limit=2)]
            self.assertEqual(sorted(comments), sorted(['before header', 'comment']))
            self.assertTrue(len(w) == 1)
            self.assertTrue(issubclass(w[-1].category, DeprecationWarning))

        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")

            comments = [str(c).strip() for c in sv.compile('').commentsiter(soup, limit=2)]
            self.assertEqual(sorted(comments), sorted(['before header', 'comment']))
            self.assertTrue(len(w) == 1)
            self.assertTrue(issubclass(w[-1].category, DeprecationWarning))



"""Test utilities."""
import unittest
import bs4
import textwrap
import soupsieve as sv


class TestCase(unittest.TestCase):
    """Test case."""

    def assert_selector(self, markup, selectors, expected_ids, namespaces={}, mode=sv.HTML5):
        """Assert selector."""

        if mode == sv.HTML:
            bs_mode = 'lxml'
        elif mode == sv.HTML5:
            bs_mode = 'html5lib'
        elif mode in (sv.XHTML, sv.XML):
            bs_mode = 'xml'
        soup = bs4.BeautifulSoup(textwrap.dedent(markup.replace('\r\n', '\n')), bs_mode)

        ids = []
        for el in sv.select(selectors, soup, namespaces=namespaces, mode=mode):
            ids.append(el.attrs['id'])
        self.assertEqual(sorted(ids), sorted(expected_ids))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Setup Soup Sieve."""

from setuptools import setup, find_packages
import os
import imp


def get_version():
    """Get version and version_info without importing the entire module."""

    path = os.path.join(os.path.dirname(__file__), 'soupsieve')
    fp, pathname, desc = imp.find_module('__meta__', [path])
    try:
        meta = imp.load_module('__meta__', fp, pathname, desc)
        return meta.__version__, meta.__version_info__._get_dev_status()
    finally:
        fp.close()


def get_requirements():
    """Get the dependencies."""

    with open("requirements/project.txt") as f:
        requirements = []
        for line in f.readlines():
            line = line.strip()
            if line and not line.startswith('#'):
                requirements.append(line)
    return requirements


def get_description():
    """Get long description."""

    with open("README.md", 'r') as f:
        desc = f.read()
    return desc


VER, DEVSTATUS = get_version()

setup(
    name='soupsieve',
    version=VER,
    python_requires=">=3.4",
    keywords='CSS HTML selector filter',
    description='A CSS4 selector implementation for Beautiful Soup.',
    long_description=get_description(),
    long_description_content_type='text/markdown',
    author='Isaac Muse',
    author_email='Isaac.Muse@gmail.com',
    url='https://github.com/facelessuser/soupsieve',
    packages=find_packages(exclude=['tests']),
    install_requires=get_requirements(),
    license='MIT License',
    classifiers=[
        'Development Status :: %s' % DEVSTATUS,
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



"""
Test selectors level 1.

```
E
E F
E, F
.class
#elementID
```
Not supported (with current opinions or plans the matter):

- `:link`: visited or un-visited links have no meaning outside a browser, just search for `:is(a,area,link)[href]` to
  target links. At most, the possibility of supporting the `:link` as an alias for `:link` may be considered in the
  future as technically, all links are un-visited in our scenario.

- `:active`: No elements in our environment can be "active", so this makes no sense in our context.
"""
from . import util
import soupsieve as sv


class TestLevel1(util.TestCase):
    """Test level 1 selectors."""

    def test_tag(self):
        """Test tag."""

        self.assert_selector(
            """
            <div>
            <p>Some text <span id="1"> in a paragraph</span>.</p>
            <a id="2" href="http://google.com">Link</a>
            </div>
            """,
            "span",
            ["1"],
            mode=sv.HTML5
        )

    def test_tags(self):
        """Test multiple selectors."""

        self.assert_selector(
            """
            <div>
            <p>Some text <span id="1"> in a paragraph</span>.
            <a id="2" href="http://google.com">Link</a>
            </p>
            </div>
            """,
            "span, a",
            ["1", "2"],
            mode=sv.HTML5
        )

    def test_child(self):
        """Test child."""

        self.assert_selector(
            """
            <div>
            <p>Some text <span id="1"> in a paragraph</span>.
            <a id="2" href="http://google.com">Link</a>
            </p>
            </div>
            """,
            "div span",
            ["1"],
            mode=sv.HTML5
        )

    def test_id(self):
        """Test ID."""

        markup = """
        <div>
        <p>Some text <span id="1"> in a paragraph</span>.
        <a id="2" href="http://google.com">Link</a>
        </p>
        </div>
        """

        self.assert_selector(
            markup,
            "#1",
            ["1"],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            "a#2",
            ["2"],
            mode=sv.HTML5
        )

    def test_class(self):
        """Test class."""

        markup = """
        <div>
        <p>Some text <span id="1" class="foo"> in a paragraph</span>.
        <a id="2" class="bar" href="http://google.com">Link</a>
        </p>
        </div>
        """

        self.assert_selector(
            markup,
            ".foo",
            ["1"],
            mode=sv.HTML5
        )

        self.assert_selector(
            markup,
            "a.bar",
            ["2"],
            mode=sv.HTML5
        )

    def test_classes(self):
        """Test classes."""

        markup = """
        <div>
        <p>Some text <span id="1" class="foo"> in a paragraph</span>.
        <a id="2" class="bar" href="http://google.com">Link</a>
        <a id="3" class="foo" href="http://google.com">Link</a>
        <a id="4" class="foo bar" href="http://google.com">Link</a>
        </p>
        </div>
        """

        self.assert_selector(
            markup,
            "a.foo.bar",
            ["4"],
            mode=sv.HTML5
        )



"""
Test selectors level 2.

```
*
:first-child
E > F
E + F
[foo]
[foo='bar']
[foo~='bar']
[foo|='en']
```

Not supported (with current opinions or plans the matter):

- `:lang(en)`: In documents, `:lang()` can take into considerations information in `meta` and other things in the
  header. At this point, there are no plans to implement this. If a reasonable proposal was introduced on how to
  support this, it may be considered.

- `:hover`: Items cannot be hovered in our environment, so this has little meaning and will not be implemented.

- `:focus`: Items cannot be focused in our environment, so this has little meaning and will not be implemented.
"""
from . import util
import soupsieve as sv


class TestLevel2(util.TestCase):
    """Test level 2 selectors."""

    def test_direct_child(self):
        """Test direct child."""

        markup = """
        <div>
        <p id="0">Some text <span id="1"> in a paragraph</span>.</p>
        <a id="2" href="http://google.com">Link</a>
        <span id="3">Direct child</span>
        <pre>
        <span id="4">Child 1</span>
        <span id="5">Child 2</span>
        <span id="6">Child 3</span>
        </pre>
        </div>
        """

        # Spaces
        self.assert_selector(
            markup,
            "div > span",
            ["3"],
            mode=sv.HTML5
        )

        # No spaces
        self.assert_selector(
            markup,
            "div>span",
            ["3"],
            mode=sv.HTML5
        )

    def test_direct_sibling(self):
        """Test direct sibling."""

        markup = """
        <div>
        <p id="0">Some text <span id="1"> in a paragraph</span>.</p>
        <a id="2" href="http://google.com">Link</a>
        <span id="3">Direct child</span>
        <pre>
        <span id="4">Child 1</span>
        <span id="5">Child 2</span>
        <span id="6">Child 3</span>
        </pre>
        </div>
        """

        # Spaces
        self.assert_selector(
            markup,
            "span + span",
            ["5", "6"],
            mode=sv.HTML5
        )

        # No spaces
        self.assert_selector(
            markup,
            "span+span",
            ["5", "6"],
            mode=sv.HTML5
        )

        # Complex
        self.assert_selector(
            markup,
            "span#4 + span#5",
            ["5"],
            mode=sv.HTML5
        )

    def test_wild_tag(self):
        """Test wild tag."""

        self.assert_selector(
            """
            <div id="div">
            <p id="0">Some text <span id="1"> in a paragraph</span>.</p>
            <a id="2" href="http://google.com">Link</a>
            <span id="3">Direct child</span>
            <pre id="pre">
            <span id="4">Child 1</span>
            <span id="5">Child 2</span>
            <span id="6">Child 3</span>
            </pre>
            </div>
            """,
            "body *",
            ["0", "1", "2", "3", "4", "5", "6", "div", "pre"],
            mode=sv.HTML5
        )

    def test_attribute(self):
        """Test attribute."""

        self.assert_selector(
            """
            <div id="div">
            <p id="0">Some text <span id="1"> in a paragraph</span>.</p>
            <a id="2" href="http://google.com">Link</a>
            <span id="3">Direct child</span>
            <pre id="pre">
            <span id="4">Child 1</span>
            <span id="5">Child 2</span>
            <span id="6">Child 3</span>
            </pre>
            </div>
            """,
            "[href]",
            ["2"],
            mode=sv.HTML5
        )

    def test_multi_attribute(self):
        """Test multiple attribute."""

        self.assert_selector(
            """
            <div id="div">
            <p id="0">Some text <span id="1"> in a paragraph</span>.</p>
            <a id="2" href="http://google.com">Link</a>
            <span id="3">Direct child</span>
            <pre id="pre">
            <span id="4" class="test">Child 1</span>
            <span id="5" class="test" data-test="test">Child 2</span>
            <span id="6">Child 3</span>
            <span id="6">Child 3</span>
            </pre>
            </div>
            """,
            "span[id].test[data-test=test]",
            ["5"],
            mode=sv.HTML5
        )

    def test_attribute_equal(self):
        """Test attribute with value that equals specified value."""

        markup = """
        <div id="div">
        <p id="0">Some text <span id="1"> in a paragraph</span>.</p>
        <a id="2" href="http://google.com">Link</a>
        <span id="3">Direct child</span>
        <pre id="pre">
        <span id="4">Child 1</span>
        <span id="5">Child 2</span>
        <span id="6">Child 3</span>
        </pre>
        </div>
        """

        # No quotes
        self.assert_selector(
            markup,
            '[id=5]',
            ["5"],
            mode=sv.HTML5
        )

        # Single quoted
        self.assert_selector(
            markup,
            "[id='5']",
            ["5"],
            mode=sv.HTML5
        )

        # Double quoted
        self.assert_selector(
            markup,
            '[id="5"]',
            ["5"],
            mode=sv.HTML5
        )

    def test_attribute_start_dash(self):
        """Test attribute whose dash separated value starts with the specified value."""

        self.assert_selector(
            """
            <div id="div">
            <p id="0" lang="en-us">Some text <span id="1"> in a paragraph</span>.</p>
            <a id="2" href="http://google.com">Link</a>
            <span id="3">Direct child</span>
            <pre id="pre">
            <span id="4">Child 1</span>
            <span id="5">Child 2</span>
            <span id="6">Child 3</span>
            </pre>
            </div>
            """,
            "[lang|=en]",
            ["0"],
            mode=sv.HTML5
        )

    def test_attribute_contains_space(self):
        """Test attribute whose space separated list contains the specified value."""

        markup = """
        <div id="div">
        <p id="0" class="test1 test2 test3">Some text <span id="1"> in a paragraph</span>.</p>
        <a id="2" href="http://google.com">Link</a>
        <span id="3">Direct child</span>
        <pre id="pre" class="test-a test-b">
        <span id="4">Child 1</span>
        <span id="5">Child 2</span>
        <span id="6">Child 3</span>
        </pre>
        </div>
        """

        # Middle of list
        self.assert_selector(
            markup,
            "[class~=test2]",
            ["0"],
            mode=sv.HTML5
        )

        # Start of list
        self.assert_selector(
            markup,
            "[class~=test-a]",
            ["pre"],
            mode=sv.HTML5
        )

        # End of list
        self.assert_selector(
            markup,
            "[class~=test-b]",
            ["pre"],
            mode=sv.HTML5
        )

    def test_first_child(self):
        """Test first child."""

        self.assert_selector(
            """
            <div id="div">
            <p id="0">Some text <span id="1"> in a paragraph</span>.</p>
            <a id="2" href="http://google.com">Link</a>
            <span id="3">Direct child</span>
            <pre id="pre">
            <span id="4">Child 1</span>
            <span id="5">Child 2</span>
            <span id="6">Child 3</span>
            </pre>
            </div>
            """,
            "span:first-child",
            ["1", "4"],
            mode=sv.HTML5
        )



"""CSS matcher."""
from . import util
import re

# Empty tag pattern (whitespace okay)
RE_NOT_EMPTY = re.compile('[^ \t\r\n]')

# Relationships
REL_PARENT = ' '
REL_CLOSE_PARENT = '>'
REL_SIBLING = '~'
REL_CLOSE_SIBLING = '+'

# Relationships for :has() (forward looking)
REL_HAS_PARENT = ': '
REL_HAS_CLOSE_PARENT = ':>'
REL_HAS_SIBLING = ':~'
REL_HAS_CLOSE_SIBLING = ':+'


class CSSMatch:
    """Perform CSS matching."""

    def __init__(self, selectors, namespaces, mode):
        """Initialize."""

        self.selectors = selectors
        self.namespaces = namespaces
        self.mode = mode

    def get_namespace(self, el):
        """Get the namespace for the element."""

        namespace = ''
        ns = el.namespace
        if ns:
            namespace = ns
        return namespace

    def supports_namespaces(self):
        """Check if namespaces are supported in the HTML type."""

        return self.mode in (util.HTML5, util.XHTML, util.XML)

    def is_xml(self):
        """Check if document is an XML type."""

        return self.mode in (util.XHTML, util.XML)

    def get_attribute(self, el, attr, prefix):
        """Get attribute from element if it exists."""

        value = None
        is_xml = self.is_xml()
        if self.supports_namespaces():
            value = None
            # If we have not defined namespaces, we can't very well find them, so don't bother trying.
            if prefix and prefix not in self.namespaces and prefix != '*':
                return None

            for k, v in el.attrs.items():
                parts = k.split(':', 1)
                if len(parts) > 1:
                    if not parts[0]:
                        a = k
                        p = ''
                    else:
                        p = parts[0]
                        a = parts[1]
                else:
                    p = ''
                    a = k
                # Can't match a prefix attribute as we haven't specified one to match
                if not prefix and p:
                    continue
                # We can't match our desired prefix attribute as the attribute doesn't have a prefix
                if prefix and not p and prefix != '*':
                    continue
                if is_xml:
                    # The prefix doesn't match
                    if prefix and p and prefix != '*' and prefix != p:
                        continue
                    # The attribute doesn't match.
                    if attr != a:
                        continue
                else:
                    # The prefix doesn't match
                    if prefix and p and prefix != '*' and util.lower(prefix) != util.lower(p):
                        continue
                    # The attribute doesn't match.
                    if util.lower(attr) != util.lower(a):
                        continue
                value = v
                break
        else:
            for k, v in el.attrs.items():
                if util.lower(attr) != util.lower(k):
                    continue
                value = v
                break
        return value

    def get_classes(self, el):
        """Get classes."""

        if self.mode not in (util.XHTML, util.XML):
            return el.attrs.get('class', [])
        else:
            return [c for c in el.attrs.get('class', '').strip().split(' ') if c]

    def match_namespace(self, el, tag):
        """Match the namespace of the element."""

        match = True
        namespace = self.get_namespace(el)
        default_namespace = self.namespaces.get('')
        # We must match the default namespace if one is not provided
        if tag.prefix is None and (default_namespace is not None and namespace != default_namespace):
            match = False
        # If we specified `|tag`, we must not have a namespace.
        elif (tag.prefix is not None and tag.prefix == '' and namespace):
            match = False
        # Verify prefix matches
        elif (
            tag.prefix and
            tag.prefix != '*' and namespace != self.namespaces.get(tag.prefix, '')
        ):
            match = False
        return match

    def match_attributes(self, el, attributes):
        """Match attributes."""

        match = True
        if attributes:
            for a in attributes:
                value = self.get_attribute(el, a.attribute, a.prefix)
                if isinstance(value, list):
                    value = ' '.join(value)
                if a.pattern is None and value is None:
                    match = False
                    break
                elif a.pattern is not None and value is None:
                    match = False
                    break
                elif a.pattern is None:
                    continue
                elif value is None or a.pattern.match(value) is None:
                    match = False
                    break
        return match

    def match_tagname(self, el, tag):
        """Match tag name."""

        return not (
            tag.name and
            tag.name not in ((util.lower(el.name) if not self.is_xml() else el.name), '*')
        )

    def match_tag(self, el, tags):
        """Match the tag."""

        has_ns = self.supports_namespaces()
        match = True
        for t in tags:
            # Verify namespace
            if has_ns and not self.match_namespace(el, t):
                match = False
                break
            if not self.match_tagname(el, t):
                match = False
                break
        return match

    def match_past_relations(self, el, relation):
        """Match past relationship."""

        found = False
        if relation.rel_type == REL_PARENT:
            parent = el.parent
            while not found and parent:
                found = self.match_selectors(parent, [relation])
                parent = parent.parent
        elif relation.rel_type == REL_CLOSE_PARENT:
            parent = el.parent
            if parent:
                found = self.match_selectors(parent, [relation])
        elif relation.rel_type == REL_SIBLING:
            sibling = el.previous_sibling
            while not found and sibling:
                if not isinstance(sibling, util.TAG):
                    sibling = sibling.previous_sibling
                    continue
                found = self.match_selectors(sibling, [relation])
                sibling = sibling.previous_sibling
        elif relation.rel_type == REL_CLOSE_SIBLING:
            sibling = el.previous_sibling
            while sibling and not isinstance(sibling, util.TAG):
                sibling = sibling.previous_sibling
            if sibling and isinstance(sibling, util.TAG):
                found = self.match_selectors(sibling, [relation])
        return found

    def match_future_child(self, parent, relation, recursive=False):
        """Match future child."""

        match = False
        for child in (parent.descendants if recursive else parent.children):
            if not isinstance(child, util.TAG):
                continue
            match = self.match_selectors(child, [relation])
            if match:
                break
        return match

    def match_future_relations(self, el, relation):
        """Match future relationship."""

        found = False
        if relation.rel_type == REL_HAS_PARENT:
            found = self.match_future_child(el, relation, True)
        elif relation.rel_type == REL_HAS_CLOSE_PARENT:
            found = self.match_future_child(el, relation)
        elif relation.rel_type == REL_HAS_SIBLING:
            sibling = el.next_sibling
            while not found and sibling:
                if not isinstance(sibling, util.TAG):
                    sibling = sibling.next_sibling
                    continue
                found = self.match_selectors(sibling, [relation])
                sibling = sibling.next_sibling
        elif relation.rel_type == REL_HAS_CLOSE_SIBLING:
            sibling = el.next_sibling
            while sibling and not isinstance(sibling, util.TAG):
                sibling = sibling.next_sibling
            if sibling and isinstance(sibling, util.TAG):
                found = self.match_selectors(sibling, [relation])
        return found

    def match_relations(self, el, relation):
        """Match relationship to other elements."""

        found = False

        if relation.rel_type.startswith(':'):
            found = self.match_future_relations(el, relation)
        else:
            found = self.match_past_relations(el, relation)

        return found

    def match_id(self, el, ids):
        """Match element's ID."""

        found = True
        for i in ids:
            if i != el.attrs.get('id', ''):
                found = False
                break
        return found

    def match_classes(self, el, classes):
        """Match element's classes."""

        current_classes = self.get_classes(el)
        found = True
        for c in classes:
            if c not in current_classes:
                found = False
                break
        return found

    def match_root(self, el):
        """Match element as root."""

        parent = el.parent
        return parent and not parent.parent

    def match_nth_tag_type(self, el, child):
        """Match tag type for `nth` matches."""

        return(
            (child.name == (util.lower(el.name) if not self.is_xml() else el.name)) and
            (self.supports_namespaces() and self.get_namespace(child) == self.get_namespace(el))
        )

    def match_nth(self, el, nth):
        """Match `nth` elements."""

        matched = True

        for n in nth:
            matched = False
            if not el.parent:
                break
            if n.selectors and not self.match_selectors(el, n.selectors):
                break
            parent = el.parent
            last = n.last
            last_index = len(parent.contents) - 1
            relative_index = 0
            a = n.a
            b = n.b
            var = n.n
            count = 0
            count_incr = 1
            factor = -1 if last else 1
            index = len(parent.contents) - 1 if last else 0
            idx = last_idx = a * count + b if var else a

            # We can only adjust bounds within a variable index
            if var:
                # Abort if our nth index is out of bounds and only getting further out of bounds as we increment.
                # Otherwise, increment to try to get in bounds.
                adjust = None
                while idx < 1 or idx > last_index:
                    diff_low = 0 - idx
                    if idx < 0:
                        if adjust is not None:
                            break
                        adjust = -1
                        count += count_incr
                        idx = last_idx = a * count + b if var else a
                        diff = 0 - idx
                        if diff >= diff_low:
                            break
                        diff_low = diff
                    diff_high = idx - last_index
                    if idx > last_index:
                        if adjust is not None:
                            break
                        adjust = 1
                        count += count_incr
                        idx = last_idx = a * count + b if var else a
                        diff = idx - last_index
                        if diff >= diff_high:
                            break
                        diff_high = diff

                # If a < 0, our count is working backwards, so floor the index by increasing the count.
                # Find the count that yields the lowest, in bound value and use that.
                # Lastly reverse count increment so that we'll increase our index.
                lowest = count
                if a < 0:
                    while idx >= 1:
                        lowest = count
                        count += count_incr
                        idx = last_idx = a * count + b if var else a
                    count_incr = -1
                count = lowest
                idx = last_idx = a * count + b if var else a

            # Evaluate elements while our calculated nth index is still in range
            while 1 <= idx <= last_index:
                child = None
                # Evaluate while our child index is still range.
                while 0 <= index <= last_index:
                    child = parent.contents[index]
                    index += factor
                    if not isinstance(child, util.TAG):
                        continue
                    # Handle `of S` in `nth-child`
                    if n.selectors and not self.match_selectors(child, n.selectors):
                        continue
                    # Handle `of-type`
                    if n.type and not self.match_nth_tag_type(el, child):
                        continue
                    relative_index += 1
                    if relative_index == idx:
                        if child is el:
                            matched = True
                        else:
                            break
                    if child is el:
                        break
                if child is el:
                    break
                last_idx = idx
                count += count_incr
                if count < 0:
                    # Count is counting down and has now ventured into invalid territory.
                    break
                idx = a * count + b if var else a
                if last_idx == idx:
                    break
            if not matched:
                break
        return matched

    def has_child(self, el):
        """Check if element has child."""

        found_child = False
        for child in el.children:
            if isinstance(child, util.CHILD):
                found_child = True
                break
        return found_child

    def match_empty(self, el, is_empty):
        """Check if element is empty (if requested)."""

        return not is_empty or (RE_NOT_EMPTY.search(el.text) is None and not self.has_child(el))

    def match_selectors(self, el, selectors):
        """Check if element matches one of the selectors."""

        match = False
        is_html = self.mode != util.XML
        for selector in selectors:
            match = selector.is_not
            # Verify tag matches
            if not self.match_tag(el, selector.tags):
                continue
            # Verify `nth` matches
            if not self.match_nth(el, selector.nth):
                continue
            if not self.match_empty(el, selector.is_empty):
                continue
            # Verify id matches
            if is_html and selector.ids and not self.match_id(el, selector.ids):
                continue
            # Verify classes match
            if is_html and selector.classes and not self.match_classes(el, selector.classes):
                continue
            # Verify attribute(s) match
            if not self.match_attributes(el, selector.attributes):
                continue
            if selector.is_root and not self.match_root(el):
                continue
            # Verify pseudo selector patterns
            if selector.selectors and not self.match_selectors(el, selector.selectors):
                continue
            # Verify relationship selectors
            if selector.relation and not self.match_relations(el, selector.relation):
                continue
            match = not selector.is_not
            break

        return match

    def match(self, el):
        """Match."""

        return isinstance(el, util.TAG) and self.match_selectors(el, self.selectors)
